#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import enum
from types import SimpleNamespace
import unittest

from balladeer.lite.entity import Entity
from balladeer.lite.performance import Performance


class Trivial(Performance):

    def do_this(self, this, text, context):
        """
        This?

        """
        yield "Yes, this."

    def do_that(self, this, text, context):
        """
        That?

        """
        yield from ["Yes.", "That."]

    def do_tother(self, this, text, context):
        """
        Or?

        """
        yield "Or,"
        yield "Maybe;"
        yield "Tother."


class MatchingTests(unittest.TestCase):

    class Location(enum.Enum):
        HERE = "here"
        THERE = "there"

    class Liquid(Entity): pass
    class Mass(Entity): pass
    class Space(Entity): pass

    def test_unpack_annotation_single_enum(self):
        rv = list(Performance.unpack_annotation("locn", MatchingTests.Location, ensemble=[]))
        self.assertTrue(rv)
        self.assertTrue(all(isinstance(i, tuple) for i in rv), rv)
        self.assertEqual(
            len([v for i in MatchingTests.Location for v in ([i.value] if isinstance(i.value, str) else i.value)]),
            len(rv),
            rv
        )

    def test_unpack_annotation_enum_list(self):
        class Season(enum.Enum):
            spring = "Spring"
            summer = "Summer"
            autumn = "Autumn"
            winter = "Winter"

        rv = list(Performance.unpack_annotation("item", [MatchingTests.Location, Season], ensemble=[]))
        self.assertTrue(rv)
        self.assertTrue(all(isinstance(i, tuple) for i in rv), rv)
        self.assertTrue(all(isinstance(i[0], str) for i in rv), rv)
        self.assertTrue(all(isinstance(i[1], enum.Enum) for i in rv), rv)
        self.assertEqual(
            len([
                v for i in list(MatchingTests.Location) + list(Season)
                for v in ([i.value] if isinstance(i.value, str) else i.value)
            ]),
            len(rv),
            rv
        )

    def test_unpack_annotation_single_class(self):
        ensemble = [MatchingTests.Liquid(), MatchingTests.Mass(), MatchingTests.Space()]
        rv = list(Performance.unpack_annotation("item", MatchingTests.Liquid, ensemble))
        self.assertTrue(rv)
        self.assertTrue(all(isinstance(i, tuple) for i in rv), rv)
        self.assertEqual(1, len(rv), rv)
        self.assertIsInstance(rv[0][1], MatchingTests.Liquid)

    def test_unpack_annotation_dataobject(self):
        ensemble = [MatchingTests.Liquid(), MatchingTests.Mass(), MatchingTests.Space()]
        rv = list(Performance.unpack_annotation("thing", [MatchingTests.Liquid, MatchingTests.Mass], ensemble))
        self.assertTrue(all(isinstance(i, tuple) for i in rv), rv)
        self.assertEqual(2, len(rv), rv)
        self.assertIsInstance(rv[0][1], MatchingTests.Liquid)
        self.assertIsInstance(rv[1][1], MatchingTests.Mass)

    def test_unpack_annotation_parent_attribute(self):
        class Season(enum.Enum):
            spring = "Spring"
            summer = "Summer"
            autumn = "Autumn"
            winter = "Winter"

            @property
            def follows(self):
                items = list(Season)
                pos = (items.index(self) - 1) % len(items)
                return items[pos]

        self.assertEqual(Season.summer, Season.autumn.follows)
        s = Season.autumn
        rv = list(Performance.unpack_annotation("item", "follows", ensemble=[], parent=s))
        self.assertTrue(rv)
        self.assertTrue(all(isinstance(i, tuple) for i in rv), rv)
        self.assertTrue(all(isinstance(i[0], str) for i in rv), rv)
        self.assertTrue(all(isinstance(i[1], enum.Enum) for i in rv), rv)
        self.assertEqual(Season.summer, rv[0][1])

    def test_unpack_annotation_parent_attribute_dotted(self):
        class Season(enum.Enum):
            spring = "Spring"
            summer = "Summer"
            autumn = "Autumn"
            winter = "Winter"

            @property
            def follows(self):
                items = list(Season)
                pos = (items.index(self) - 1) % len(items)
                return items[pos]

        self.assertEqual(Season.summer, Season.autumn.follows)
        obj = SimpleNamespace(s=Season.autumn)
        rv = list(Performance.unpack_annotation("item", "s.follows", ensemble=[], parent=obj))
        self.assertTrue(rv)
        self.assertTrue(all(isinstance(i, tuple) for i in rv), rv)
        self.assertTrue(all(isinstance(i[0], str) for i in rv), rv)
        self.assertTrue(all(isinstance(i[1], enum.Enum) for i in rv), rv)
        self.assertEqual(Season.summer, rv[0][1])

    def test_unpack_annotation_parent_attribute_mixed(self):
        obj = SimpleNamespace(one=SimpleNamespace(two={"three": 3}))
        rv = list(Performance.unpack_annotation("item", "one.two[three]", ensemble=[], parent=obj))
        self.assertEqual(("item", 3), rv[0])

    def test_expand_commands_no_preserver(self):

        thing = Entity(name="thing")

        def func(obj: Entity):
            """
            pick up a {obj.name}
            """

        rv = dict(Performance.expand_commands(func, ensemble=[thing]))
        self.assertIn("pick up thing", rv)

    def test_expand_commands_sequence_with_preserver(self):

        thing = Entity(names=["thing"])

        def func(obj: Entity):
            """
            pick up a {obj.names[0]}.
            """

        rv = dict(Performance.expand_commands(func, ensemble=[thing]))
        self.assertIn("pick up a thing", rv)

    def test_expand_commands_sequence_no_preserver(self):

        thing = Entity(names=["thing"])

        def func(obj: Entity):
            """
            pick up a {obj.names[0]}
            """

        rv = dict(Performance.expand_commands(func, ensemble=[thing]))
        self.assertIn("pick up thing", rv)

    def test_expand_commands_sequence_synonyms_no_preserver(self):

        thing = Entity(names=["thing", "doobrey"])
        idea = Entity(names=["idea"])

        def func(obj: [object, Entity]):
            """
            pick up a {obj.names[0]} | grab a {obj.names[0]}
            pick up a {obj.names[1]} | grab a {obj.names[1]}
            """

        rv = dict(Performance.expand_commands(func, ensemble=[idea, thing, idea]))
        self.assertIn("pick up thing", rv)
        self.assertIn("pick up doobrey", rv)
        self.assertIn("grab thing", rv)
        self.assertIn("grab doobrey", rv)

    def test_expand_commands_with_preserver(self):

        thing = Entity(name="thing")

        def func(obj: Entity):
            """
            pick up a {obj.name}.
            """

        rv = dict(Performance.expand_commands(func, ensemble=[thing]))
        self.assertIn("pick up a thing", rv)

    def test_expand_commands_discrimminator(self):

        ensemble = [
            Entity(names=["thing"], aspect="blue"),
            Entity(names=["thing"], aspect="red"),
        ]

        def func(obj: Entity):
            """
            pick up a {obj.aspect} {obj.names[0]}
            """

        rv = dict(Performance.expand_commands(func, ensemble=ensemble))
        self.assertIn("pick up blue thing", rv)
        self.assertEqual("blue", rv["pick up blue thing"][1]["obj"].aspect)
        self.assertIn("pick up red thing", rv)
        self.assertEqual("red", rv["pick up red thing"][1]["obj"].aspect)


class PerformanceMatchTests(unittest.TestCase):

    def setUp(self):
        self.performance = Trivial()

    def test_do_that(self):
        fn, args, kwargs = next(self.performance.actions("that?"))
        self.assertEqual(self.performance.do_that, fn)
        self.assertEqual(["that?", None], args)
        self.assertFalse(kwargs)

    def test_mismatch(self):
        cmd = "release the frog"
        fn, args, kwargs = next(self.performance.actions(cmd))
        self.assertIs(None, fn)
        self.assertEqual([cmd, None], args)
        self.assertFalse(kwargs)


class PerformanceTests(unittest.TestCase):

    def setUp(self):
        self.performance = Trivial()

    def test_do_that(self):
        fn, args, kwargs = next(self.performance.actions("that?"))
        self.assertEqual(self.performance.do_that, fn)
        self.assertEqual(["that?", None], args)
        self.assertFalse(kwargs)

        fn, args, kwargs = self.performance.pick([(fn, args, kwargs)])
        self.assertEqual(self.performance.do_that, fn)
        self.assertEqual(["that?", None], args)

    def test_do_this(self):
        fn, args, kwargs = self.performance.pick(self.performance.actions("That?"))
        data = self.performance(fn, *args, **kwargs)

        fn, args, kwargs = self.performance.pick(self.performance.actions("This?"))
        data = self.performance(fn, *args, **kwargs)
        self.assertEqual("Yes, this.", "\n".join(data))

    def test_do_tother(self):
        fn, args, kwargs = self.performance.pick(self.performance.actions("That?"))
        data = self.performance(fn, *args, **kwargs)

        fn, args, kwargs = self.performance.pick(self.performance.actions("This?"))
        data = self.performance(fn, *args, **kwargs)

        fn, args, kwargs = self.performance.pick(self.performance.actions("Or?"))
        data = "\n".join(self.performance(fn, *args, **kwargs))

        self.assertEqual("Or,\nMaybe;\nTother.", data)



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import functools
import re
import textwrap

from speechmark import SpeechMark


class Speech(str):
    processor = SpeechMark()
    tag_matcher = re.compile("<[^>]+?>")

    @functools.cache
    def trim(self) -> str:
        return textwrap.dedent(self).strip()

    @functools.cached_property
    def tags(self) -> str:
        return self.processor.loads(self.trim())

    @functools.cached_property
    def lines(self) -> list[str]:
        text = self.tag_matcher.sub("", self.tags)
        return list(filter(None, (i.strip() for i in text.splitlines())))

    @functools.cached_property
    def words(self) -> list[str]:
        return " ".join(self.lines).split(" ")


class Prologue(Speech):
    "Speech before a scene."


class Dialogue(Speech):
    "Speech during a scene."


class Epilogue(Speech):
    "Speech after a scene."



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy
import enum
import textwrap
import unittest
import tomllib

from balladeer.lite.director import Director
from balladeer.lite.entity import Entity
from balladeer.lite.speech import Speech
from balladeer.lite.speech import Prologue
from balladeer.lite.speech import Dialogue
from balladeer.lite.speech import Epilogue
from balladeer.lite.types import State


class EditTests(unittest.TestCase):
    def test_word_counter(self):
        text = textwrap.dedent("""
        <STAFF.proposing#3> What will you have, sir? The special is fish today.

            1. Order the Beef Wellington
            2. Go for the Shepherd's Pie
            3. Try the Dover Sole
        """).strip()
        speech = Speech(text)

        director = Director()
        self.assertEqual(5, len(speech.lines))
        self.assertEqual(24, len(speech.words))

        rv = "\n".join(director.edit(speech))
        self.assertFalse("<p>\n\n</p>" in rv, rv)

    def test_animate_cite_tag(self):
        text = textwrap.dedent("""
        <FIGHTER_1>

            I don't like the way you use me, {FIGHTER_2.name}!

        """).strip()
        director = Director()
        selection = {
            "FIGHTER_1": Entity(name="Biffy"),
            "FIGHTER_2": Entity(name="Bashy"),
        }

        speech = Speech(text)
        edit = "\n".join(director.edit(speech, selection))
        match = director.cite_matcher.search(edit)
        self.assertTrue(match)
        self.assertIn("animation-delay", match.group())
        self.assertIn("animation-duration", match.group())

    def test_edit_single_blocks(self):
        text = textwrap.dedent("""
        <FIGHTER_1>

            I don't like the way you use me, {FIGHTER_2.name}!

        """).strip()
        director = Director()
        selection = {
            "FIGHTER_1": Entity(name="Biffy"),
            "FIGHTER_2": Entity(name="Bashy"),
        }

        speech = Speech(text)
        edit = "\n".join(director.edit(speech, selection))
        self.assertIn('data-role="FIGHTER_1"', edit)
        self.assertIn('data-entity="Biffy"', edit, speech)
        self.assertIn(">Biffy</cite>", edit)
        self.assertIn("Bashy!", edit)

    def test_edit_extended_characters_unescaped(self):
        text = textwrap.dedent("""
        <FIGHTER_1>

            I don't like the way you use me, {FIGHTER_2.name}!

        """).strip()
        director = Director()
        selection = {
            "FIGHTER_1": Entity(name="Bîffy"),
            "FIGHTER_2": Entity(name="Båshy"),
        }

        speech = Speech(text)
        edit = "\n".join(director.edit(speech, selection))
        self.assertIn('data-role="FIGHTER_1"', edit)
        self.assertIn('data-entity="Bîffy"', edit)
        self.assertIn("B&icirc;ffy</cite>", edit)
        self.assertIn("Båshy!", edit)

    def test_edit_extended_characters_escaped(self):
        text = textwrap.dedent("""
        <FIGHTER_1>

            I don't like the way you use me, {FIGHTER_2.name!a}!

        """).strip()
        director = Director()
        selection = {
            "FIGHTER_1": Entity(name="Bîffy"),
            "FIGHTER_2": Entity(name="Båshy"),
        }

        speech = Speech(text)
        edit = "\n".join(director.edit(speech, selection))
        self.assertIn('data-role="FIGHTER_1"', edit)
        self.assertIn('data-entity="Bîffy"', edit)
        self.assertIn("B&icirc;ffy</cite>", edit)
        self.assertIn("B&aring;shy!", edit)

    def test_edit_multiple_blocks(self):
        text = textwrap.dedent("""
        <WEAPON.attacking@FIGHTER_2:shouts/slapwhack>

            _Whack!_

        <FIGHTER_2>

            Uuurrggh!

        """).strip()
        director = Director()
        selection = {
            "WEAPON": Entity(name="Rusty"),
            "FIGHTER_2": Entity(name="Bashy"),
        }

        speech = Speech(text)
        edit = "\n".join(director.edit(speech, selection))
        self.assertIn('data-role="WEAPON"', edit)
        self.assertIn('data-entity="Rusty"', edit, speech)
        self.assertIn(">Rusty</cite>", edit)

        self.assertIn('data-role="FIGHTER_2"', edit)
        self.assertIn('data-entity="Bashy"', edit, speech)
        self.assertIn(">Bashy</cite>", edit)


class IntegerConditionTests(unittest.TestCase):
    def test_single_state_selector(self):
        ensemble = [Entity().set_state(1)]
        content = textwrap.dedent("""
        [ENTITY]
        state = 1

        [[_]]

        """).strip()

        d = Director()
        scene = tomllib.loads(content)
        conditions = [dict(d.specify_conditions(shot)) for shot in scene.get(d.shot_key)]

        specs = d.specifications(scene)

        roles = dict(d.roles(specs, ensemble))
        self.assertTrue(d.allows(conditions[0], roles))

    def test_single_state_condition(self):
        ensemble = [Entity().set_state(1)]
        content = textwrap.dedent("""
        [ENTITY]

        [[_]]
        if.ENTITY.state = 0

        [[_]]
        if.ENTITY.state = 1

        """).strip()

        d = Director()
        scene = tomllib.loads(content)
        conditions = [dict(d.specify_conditions(shot)) for shot in scene.get(d.shot_key)]
        self.assertEqual(2, len(conditions))

        specs = d.specifications(scene)

        roles = dict(d.roles(specs, ensemble))
        self.assertFalse(d.allows(conditions[0], roles))
        self.assertTrue(d.allows(conditions[1], roles))

    def test_multi_state_selector(self):
        ensemble = [Entity().set_state(1)]
        content = textwrap.dedent("""
        [ENTITY]
        states.int = [0, 2, 4, 6]

        [[_]]

        """).strip()

        d = Director()
        scene = tomllib.loads(content)
        conditions = [dict(d.specify_conditions(shot)) for shot in scene.get(d.shot_key)]

        specs = d.specifications(scene)

        roles = dict(d.roles(specs, ensemble))
        self.assertTrue(d.allows(conditions[0], roles))

    def test_multi_state_condition(self):
        ensemble = [Entity().set_state(1)]
        content = textwrap.dedent("""
        [ENTITY]

        [[_]]
        if.ENTITY.states.int = [0, 2, 4, 6]

        [[_]]
        if.ENTITY.states.int = [1, 3, 5]

        """).strip()

        d = Director()
        scene = tomllib.loads(content)
        conditions = [dict(d.specify_conditions(shot)) for shot in scene.get(d.shot_key)]
        self.assertEqual(2, len(conditions))

        specs = d.specifications(scene)

        roles = dict(d.roles(specs, ensemble))
        self.assertFalse(d.allows(conditions[0], roles))
        self.assertTrue(d.allows(conditions[1], roles))

    def test_multi_implicit_state_selector(self):
        ensemble = [Entity().set_state(1)]
        content = textwrap.dedent("""
        [ENTITY]
        state = [0, 2, 4, 6]

        [[_]]

        """).strip()

        d = Director()
        scene = tomllib.loads(content)
        conditions = [dict(d.specify_conditions(shot)) for shot in scene.get(d.shot_key)]

        specs = d.specifications(scene)

        roles = dict(d.roles(specs, ensemble))
        self.assertTrue(d.allows(conditions[0], roles))

    def test_multi_implicit_state_condition(self):
        ensemble = [Entity().set_state(1)]
        content = textwrap.dedent("""
        [ENTITY]

        [[_]]
        if.ENTITY.state = [0, 2, 4, 6]

        [[_]]
        if.ENTITY.state = [1, 3, 5]

        """).strip()

        d = Director()
        scene = tomllib.loads(content)
        conditions = [dict(d.specify_conditions(shot)) for shot in scene.get(d.shot_key)]
        self.assertEqual(2, len(conditions))

        specs = d.specifications(scene)

        roles = dict(d.roles(specs, ensemble))
        self.assertFalse(d.allows(conditions[0], roles))
        self.assertTrue(d.allows(conditions[1], roles))


class EnumConditionTests(unittest.TestCase):
    class Rain(Entity):
        pass

    class Sleet(Entity):
        pass

    class Snow(Entity):
        pass

    class Weather(State, enum.Enum):
        quiet = 0
        stormy = 1

    content = textwrap.dedent("""
        # Metadata at the top
        author = "tundish"

        [WEATHER]
        types   =   ["Rain", "Snow"]
        states.Weather = ["stormy"]

        [[_]]
        s="Outside."

        [[_]]
        # Intense
        if.WEATHER.state = "Weather.stormy"

        s="<WEATHER> It's stormy!"

        [[_]]
        # Hushed
        if.WEATHER.state = "Weather.quiet"

        s="<WEATHER> It's quiet."

        [[_]]
        # Snow storm
        if.WEATHER.type = "Snow"

        s="<WEATHER> Flurry, flurry."

        [[_]]
        # Rainfall
        if.WEATHER.type = "Rain"

        s="<WEATHER> Pitter patter."

        """).strip()

    def setUp(self):
        self.ensemble = [
            self.Rain().set_state(self.Weather.stormy),
            self.Sleet().set_state(self.Weather.stormy),
            self.Snow().set_state(self.Weather.quiet),
        ]

    def test_condition_evaluation_one(self):
        d = Director()
        scene = tomllib.loads(self.content)
        conditions = [dict(d.specify_conditions(shot)) for shot in scene.get(d.shot_key)]
        self.assertEqual(5, len(conditions))

        specs = d.specifications(scene)

        roles = dict(d.roles(specs, self.ensemble))
        self.assertTrue(d.allows(conditions[0], roles))
        self.assertTrue(d.allows(conditions[1], roles))
        self.assertFalse(d.allows(conditions[2], roles))
        self.assertFalse(d.allows(conditions[3], roles))
        self.assertTrue(d.allows(conditions[4], roles))

    def test_condition_evaluation_two(self):
        self.ensemble[0].set_state(self.Weather.quiet)
        self.ensemble[1].set_state(self.Weather.stormy)
        self.ensemble[2].set_state(self.Weather.stormy)

        d = Director()
        scene = tomllib.loads(self.content)
        conditions = [dict(d.specify_conditions(shot)) for shot in scene.get(d.shot_key)]
        self.assertEqual(5, len(conditions))

        specs = d.specifications(scene)

        self.assertEqual(3, len(self.ensemble))
        roles = dict(d.roles(specs, self.ensemble))
        self.assertTrue(roles, (self.ensemble, specs))
        self.assertTrue(d.allows(conditions[0], roles))
        self.assertTrue(d.allows(conditions[1], roles))
        self.assertFalse(d.allows(conditions[2], roles))
        self.assertTrue(d.allows(conditions[3], roles))
        self.assertFalse(d.allows(conditions[4], roles))

    def test_guard_conditions_single_state(self):
        content = textwrap.dedent("""
        [[_]]
        if.WEATHER.state = "Weather.stormy"

        """).strip()
        scene = tomllib.loads(content)
        director = Director()
        shot = next(iter(scene.get(director.shot_key)))
        conditions = dict(director.specify_conditions(shot))
        self.assertIn("WEATHER", conditions)
        roles, states, types = conditions["WEATHER"]
        self.assertIn("Weather", states)
        self.assertIn("stormy", states["Weather"])

    def test_guard_conditions_multiple_states(self):
        content = textwrap.dedent("""
        [[_]]
        if.WEATHER.states.Weather = ["stormy", "misty"]

        """).strip()
        scene = tomllib.loads(content)
        director = Director()
        shot = next(iter(scene.get(director.shot_key)))
        conditions = dict(director.specify_conditions(shot))
        self.assertIn("WEATHER", conditions)
        roles, states, types = conditions["WEATHER"]
        self.assertIn("Weather", states)
        self.assertIn("stormy", states["Weather"])
        self.assertIn("misty", states["Weather"])


class RoleTests(unittest.TestCase):
    @enum.unique
    class Aggression(State, enum.Enum):
        calm = 0
        angry = 1

    @enum.unique
    class Contentment(State, enum.Enum):
        sad = 0
        happy = 1

    @enum.unique
    class Location(State, enum.Enum):
        pub = 0
        pub_bar = 1
        pub_carpark = 2
        pub_snug = 3
        pub_toilets = 4

    def setUp(self):
        self.ensemble = [
            Entity(name="Biffy", types={"Animal", "Canine"}),
            Entity(name="Bashy", types={"Animal", "Feline"}),
            Entity(name="Rusty", type="Weapon"),
        ]

    def test_rank_constraints(self):
        roles = {
            "A": {},
            "B": {
                "state": "Switched.on",
            },
            "C": {
                "states": {
                    "Switched": ["on", "off"],
                }
            },
            "D": {
                "states": {
                    "Switched": ["on"],
                    "Spinning": ["clockwise"],
                },
            },
            "E": {
                "state": "Switched.on",
                "type": "Torch",
            },
            "F": {
                "state": "Switched.on",
                "types": ["Torch", "Desklight"],
            },
            "G": {
                "state": "Switched.on",
                "types": ["Desklight"],
            },
            "H": {
                "state": "Switched.on",
                "types": ["Torch", "Desklight"],
                "roles": ["B", "D"],
            },
        }
        ranks = [0, 1, 0.5, 2, 2, 3, 2, 5]
        d = Director()
        for n, (k, v) in enumerate(roles.items()):
            with self.subTest(role=k, spec=v):
                rank = d.rank_constraints(v)
                self.assertEqual(ranks[n], rank, d.specify_role(v))

    def test_role_with_single_required_state(self):
        text = textwrap.dedent("""
        [FIGHTER_1]
        states.Aggression = ["angry"]

        [FIGHTER_2]
        state = "Contentment.sad"

        [WEAPON]
        # A weapon which makes a noise in use.
        """)
        entities = {i.name: i for i in self.ensemble}
        entities["Biffy"].set_state(RoleTests.Contentment.sad)
        self.assertEqual(
            RoleTests.Contentment.sad,
            entities["Biffy"].get_state(RoleTests.Contentment),
        )
        entities["Bashy"].set_state(RoleTests.Aggression.angry)
        self.assertEqual(
            RoleTests.Aggression.angry,
            entities["Bashy"].get_state(RoleTests.Aggression),
        )
        scene = tomllib.loads(text)
        self.assertIsInstance(scene, dict)

        director = Director()
        rv = dict(director.roles(scene, self.ensemble))
        self.assertEqual(3, len(rv), rv)
        self.assertEqual(entities["Bashy"], rv["FIGHTER_1"])
        self.assertEqual(entities["Biffy"], rv["FIGHTER_2"])
        self.assertEqual(entities["Rusty"], rv["WEAPON"])

    def test_role_with_multiple_states(self):
        text = textwrap.dedent("""
        [FIGHTER_1]
        states.Location = ["pub_bar", "pub_toilets"]

        [FIGHTER_2]
        state = "Location.pub_bar"

        [WEAPON]
        # A weapon which makes a noise in use.
        """).strip()
        entities = {i.name: i for i in self.ensemble}
        entities["Biffy"].set_state(RoleTests.Location.pub_bar)
        self.assertEqual(
            RoleTests.Location.pub_bar, entities["Biffy"].get_state(RoleTests.Location)
        )
        entities["Bashy"].set_state(RoleTests.Location.pub_toilets)
        self.assertEqual(
            RoleTests.Location.pub_toilets,
            entities["Bashy"].get_state(RoleTests.Location),
        )

        scene = tomllib.loads(text)
        self.assertIsInstance(scene, dict)

        director = Director()
        rv = dict(director.roles(scene, self.ensemble))
        self.assertEqual(3, len(rv), rv)
        self.assertEqual(entities["Bashy"], rv["FIGHTER_1"])
        self.assertEqual(entities["Biffy"], rv["FIGHTER_2"])
        self.assertEqual(entities["Rusty"], rv["WEAPON"])

    def test_roles_are_greedy(self):
        text = textwrap.dedent("""
        [CHARACTER_1]
        roles = ["CHARACTER_2"]

        [CHARACTER_2]

        """).strip()
        scene = tomllib.loads(text)
        self.assertIsInstance(scene, dict)

        entities = {i.name: i for i in self.ensemble}

        director = Director()
        rv = dict(director.roles(scene, self.ensemble))
        self.assertEqual(2, len(rv), rv)
        self.assertEqual(entities["Biffy"], rv["CHARACTER_1"])
        self.assertEqual(entities["Biffy"], rv["CHARACTER_2"])


class ParametersTests(unittest.TestCase):
    def test_paragraph_reveal(self):
        text = textwrap.dedent("""
            <?pause=3&dwell=0.4>

            Above, there is the sound of footsteps.

            Of snagging on a threadbare carpet.

            Then shuffling down the ancient stairs.

            """).strip()

        speech = Speech(text)

        d = Director()
        rv = "\n".join(d.edit(speech))
        self.assertEqual(3, rv.count("animation-delay"), rv)
        self.assertEqual(3, rv.count("animation-duration"), rv)

    def test_offer_empty_before_param(self):
        text = textwrap.dedent("""
            <?offer=&pause=0.5>

            Which way? Take your time.

            1. Go North
            2. Go East

            """).strip()

        speech = Speech(text)

        d = Director()
        rv = "\n".join(d.edit(speech))
        self.assertIn("offer", d.notes[(None, 0)])
        self.assertEqual(None, d.notes[(None, 0)]["offer"])
        self.assertIn("delay", d.notes[(None, 0)])
        self.assertAlmostEqual(2.4, d.notes[(None, 0)]["delay"])

    def test_offer_negative(self):
        text = textwrap.dedent("""
            <?offer=1>
            <?offer=-1>

            Which way? Take your time.

            1. Go North
            2. Go East

            """).strip()

        speech = Speech(text)

        d = Director()
        rv = "\n".join(d.edit(speech))
        self.assertIn("offer", d.notes[(None, 0)])
        self.assertEqual(None, d.notes[(None, 0)]["offer"])
        self.assertIn("delay", d.notes[(None, 0)])
        self.assertAlmostEqual(3.9, d.notes[(None, 0)]["delay"])

    def test_offer_value(self):
        text = textwrap.dedent("""
            <?offer=12#2>

            Which way? You have only a few seconds to decide.

            1. Go North
            2. Go East (default)

            """).strip()

        speech = Speech(text)

        d = Director()
        rv = "\n".join(d.edit(speech))
        self.assertIn("offer", d.notes[(None, 0)])
        self.assertEqual(12, d.notes[(None, 0)]["offer"])
        self.assertIn("option", d.notes[(None, 0)])
        self.assertEqual(2, d.notes[(None, 0)]["option"])


class LoopTests(unittest.TestCase):
    def setUp(self):
        self.script = textwrap.dedent("""
        [GUEST]

        [[_]]
        s='''
        <GUEST#!>

            + This, or
            + This, or
            + This, or
        '''
        """).strip()
        self.ensemble = [Entity()]

    def test_attr_matcher(self):
        text = '<cite data-role="GUEST" data-fragments="#!">'
        d = Director()
        attrs = d.attributes(text)
        self.assertIn("role", attrs)
        self.assertEqual("GUEST", attrs["role"], attrs)
        self.assertEqual("#!", attrs["fragments"])

    def test_bang_loop(self):
        d = Director()

        scene = tomllib.loads(self.script)
        specs = d.specifications(scene)
        roles = dict(d.roles(specs, self.ensemble))

        shot = next(iter(scene.get(d.shot_key, [])))
        text = shot.get(d.dialogue_key, "")

        speech = Speech(text)

        rv = "\n".join(d.edit(speech, roles))
        self.assertEqual(0, rv.count("<li>"), rv)
        self.assertEqual(0, rv.count("</li>"))
        self.assertEqual(1, rv.count("This, or"))
        self.assertEqual(1, d.counts[(None, 0)])


class DirectiveTests(unittest.TestCase):
    def test_directive_handling(self):
        script = textwrap.dedent("""
            [GUEST]

            [STAFF]

            [PHONE]

            [[_]]

            s='''
            <PHONE.announcing@GUEST,STAFF> Ring riiing!
            '''
            """).strip()

        scene = tomllib.loads(script)
        ensemble = [Entity(), Entity(), Entity()]

        d = Director()
        specs = d.specifications(scene)
        roles = dict(d.roles(specs, ensemble))

        shot = next(iter(scene.get(d.shot_key, [])))
        text = shot.get(d.dialogue_key, "")
        speech = Speech(text)

        rv = "\n".join(d.edit(speech, roles))

        self.assertIn((None, 0), d.notes)
        self.assertEqual(1.2, d.notes[(None, 0)]["delay"])

        self.assertIn("directives", d.notes[(None, 0)])
        self.assertEqual(
            [("announcing", roles["PHONE"], (roles["GUEST"], roles["STAFF"]))],
            d.notes[(None, 0)]["directives"],
        )


class ModeTests(unittest.TestCase):
    def test_audio_media_in_multiple_blocks(self):
        text = textwrap.dedent("""
        <WEAPON.attacking@FIGHTER_2:shouts/slapwhack>

            _Whack!_

        <FIGHTER_2>

            Uuurrggh!

        """).strip()
        d = Director()
        selection = {
            "WEAPON": Entity(name="Rusty"),
            "FIGHTER_2": Entity(name="Bashy"),
        }

        speech = Speech(text)
        edit = "\n".join(d.edit(speech, selection))
        self.assertIn("media", d.notes[(None, 0)])
        self.assertEqual(["slapwhack"], d.notes[(None, 0)]["media"])


class RewriteTests(unittest.TestCase):

    def setUp(self):
        self.roles = {
            "GUEST": Entity(name="Alan"),
            "STAFF": Entity(name="Beth"),
        }

    def test_empty_rewrite(self):
        d = Director()
        rv = list(d.rewrite())
        self.assertFalse(rv)

    def test_speech_rewrite(self):
        d = Director()
        rv = list(
            d.rewrite(roles=self.roles, speech=[Prologue("<GUEST> I'd like to order a taxi.")])
        )
        self.assertEqual(1, len(rv))
        self.assertIn('data-entity="Alan"', rv[0])
        self.assertIn('>Alan</cite>', rv[0])

    def test_multi_speech_rewrite(self):
        d = Director()
        rv = list(
            d.rewrite(
                roles=self.roles,
                speech=[
                    Dialogue("<GUEST> I'd like to order a taxi, please."),
                    Epilogue("<STAFF> Certainly. Going to?"),
                    Prologue("<STAFF> Can I help you, sir?"),
                ]
            )
        )
        self.assertEqual(3, len(rv))
        self.assertIn('data-entity="Beth"', rv[0])
        self.assertIn('>Beth</cite>', rv[0])
        self.assertIn("help", rv[0])

        self.assertIn('data-entity="Alan"', rv[1])
        self.assertIn('>Alan</cite>', rv[1])
        self.assertIn("taxi", rv[1])

        self.assertIn('data-entity="Beth"', rv[2])
        self.assertIn('>Beth</cite>', rv[2])
        self.assertIn("Going", rv[2])



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from collections import namedtuple
import importlib.resources
import mimetypes
import tomllib
import xml.etree.ElementTree as ET


class Loader:
    Asset = namedtuple(
        "Asset", ["resource", "path", "type", "stats"], defaults=[None, None]
    )
    Scene = namedtuple(
        "Scene",
        ["text", "tables", "resource", "path", "stats"],
        defaults=[None, None, None],
    )

    @staticmethod
    def discover(package, resource=".", suffixes=[".scene.toml"]):
        for path in importlib.resources.files(package).joinpath(resource).iterdir():
            typ, _ = mimetypes.guess_type(path)
            if typ and typ != "text/x-python":
                with importlib.resources.as_file(path) as f:
                    yield Loader.Asset(resource, path, typ, f.stat())

            if "".join(path.suffixes) in suffixes:
                with importlib.resources.as_file(path) as f:
                    text = f.read_text(encoding="utf8")
                    yield Loader.read(text, resource=resource, path=path)

    @staticmethod
    def read(text: str, resource="", path=None):
        try:
            tables = tomllib.loads(text)
            stats = {}
        except tomllib.TOMLDecodeError as e:
            tables = None
            stats = e
        return Loader.Scene(text, tables, resource, path, stats)



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import enum
from collections import defaultdict


class Page:

    themes = {
        "default": {
            "ballad-ink-gravity": "hsl(293.33, 96.92%, 12.75%)",
            "ballad-ink-shadows": "hsl(202.86, 100%, 4.12%)",
            "ballad-ink-lolight": "hsl(203.39, 96.72%, 11.96%)",
            "ballad-ink-midtone": "hsl(203.39, 96.72%, 11.96%)",
            "ballad-ink-hilight": "hsl(203.06, 97.3%, 56.47%)",
            "ballad-ink-washout": "hsl(50, 0%, 100%, 1.0)",
            "ballad-ink-glamour": "hsl(353.33, 96.92%, 12.75%)",
        },
    }

    @enum.unique
    class Zone(enum.Enum):
        xml = enum.auto()
        doc = enum.auto()
        html = enum.auto()
        head = enum.auto()
        title = enum.auto()
        rdf = enum.auto()
        meta = enum.auto()
        link = enum.auto()
        css = enum.auto()
        theme = enum.auto()
        body = enum.auto()
        style = enum.auto()
        app = enum.auto()
        nav = enum.auto()
        main = enum.auto()
        asides = enum.auto()
        inputs = enum.auto()
        svg = enum.auto()
        iframe = enum.auto()
        script = enum.auto()
        end = enum.auto()

    def __init__(self, zone=Zone):
        self.zone = zone
        self.structure = self.setup(zone)

    def setup(self, zone):
        rv = {z: list() for z in zone}
        rv[zone.doc].append("<!DOCTYPE html>")
        rv[zone.html].append("<html>")
        rv[zone.head].append("<head>")
        rv[zone.body].extend(["</head>", "<body>"])
        # Sort links by type, eg: css, js, font, etc
        # <link
        #   rel="preload"
        #   href="fonts/zantroke-webfont.woff2"
        #   as="font"
        #   type="font/woff2"
        #   crossorigin />

        # NB: Prefetch gets resources for the next page.
        # Stateful Presenter needs lookahead.
        rv[zone.end].extend(["</body>", "</html>"])
        return rv

    def paste(self, zone, *args):
        self.structure[zone].extend(filter(None, args))
        return self

    @property
    def html(self):
        return "\n".join(
            gen if isinstance(gen, str) else "\n".join(gen)
            for seq in self.structure.values()
            for gen in seq
        )


class State:
    @classmethod
    def factory(cls, name=None, **kwargs):
        return cls[name]


class Fruition(State, enum.IntEnum):
    """
    Adapted from 'Understanding Computers and Cognition'
    by Terry Winograd and Fernando Flores,
    fig 5.1: The basic conversation for action.

    """
    inception = 1
    elaboration = 2
    construction = 3
    transition = 4
    completion = 5
    discussion = 6
    defaulted = 7
    withdrawn = 8
    cancelled = 9


class Grouping(defaultdict):

    @classmethod
    def typewise(cls, items):
        rv = cls(list)
        for i in items:
            rv[type(i)].append(i)
            try:
                for t in i.types:
                    rv[t].append(i)
            except AttributeError:
                try:
                    rv[i.type].append(i)
                except AttributeError:
                    pass
        return rv

    @property
    def all(self):
        return [i for s in self.values() for i in s]

    @property
    def each(self):
        return list(set(self.all))




import datetime
import importlib.metadata

try:
    import tomllib
except ImportError:
    import tomli as tomllib


try:
    __version__ = importlib.metadata.version("balladeer")
except importlib.metadata.PackageNotFoundError:
    __version__ = datetime.date.today().strftime("%Y.%m.%d") + "+local_repository"

from balladeer.classic.drama import Drama
from balladeer.classic.gesture import Gesture
from balladeer.classic.gesture import Hand
from balladeer.classic.gesture import Head
from balladeer.classic.speech import Article
from balladeer.classic.speech import Name
from balladeer.classic.speech import Phrase
from balladeer.classic.speech import Pronoun
from balladeer.classic.speech import Verb
from balladeer.classic.story import Story
from balladeer.classic.types import Fruition
from balladeer.classic.types import Grouping
from balladeer.classic.types import Named
from balladeer.classic.types import World



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2022 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import copy
import enum
import operator
import sys
import textwrap
import unittest
import uuid

from balladeer.lite.entity import Entity
from balladeer.lite.types import State

from balladeer.lite.loader import Loader


class SceneTests(unittest.TestCase):
    def test_one_scene(self):
        content = textwrap.dedent("""
            [[_]]

            s='''
            Text
            '''
        """)
        scene = Loader.read(content)
        self.assertIsInstance(scene, Loader.Scene)
        self.assertEqual(1, len(scene.tables.get("_")))

        self.assertEqual("Text\n", scene.tables["_"][0]["s"])

    def test_multi_scene(self):
        content = textwrap.dedent("""
            [[_]]

            s='''
            Text
            '''

            [[_]]

            s='''
            Text
            '''
        """)
        scene = Loader.read(content)
        self.assertIsInstance(scene, Loader.Scene)
        self.assertEqual(2, len(scene.tables.get("_")))

        self.assertEqual("Text\n", scene.tables["_"][0]["s"])
        self.assertEqual("Text\n", scene.tables["_"][1]["s"])



#!/usr/bin/env python3
#   encoding: utf-8

# This is part of the Balladeer library.
# Copyright (C) 2023 D E Haynes

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
from collections.abc import Generator
import json
import pathlib
import textwrap
import warnings

import hypercorn
from hypercorn.asyncio import serve

from starlette.applications import Starlette
from starlette.endpoints import HTTPEndpoint
from starlette.responses import HTMLResponse
from starlette.responses import JSONResponse
from starlette.responses import PlainTextResponse
from starlette.responses import RedirectResponse
from starlette.routing import Mount
from starlette.routing import Route
from starlette.staticfiles import StaticFiles

import balladeer
from balladeer.lite.entity import Entity
from balladeer.lite.loader import Loader
from balladeer.lite.story import StoryBuilder
from balladeer.lite.types import Grouping
from balladeer.lite.types import Page


class About(HTTPEndpoint):
    async def get(self, request):
        return PlainTextResponse(
            "\n".join(
                (
                    f"Balladeer {balladeer.__version__}",
                    "Example 10",
                )
            )
        )


class Home(HTTPEndpoint):
    meta = textwrap.dedent(
        """
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <meta http-equiv="X-UA-Compatible" content="ie=edge">
    """
    ).strip()

    body = textwrap.dedent(
        """
    <form role="form" action="/sessions" method="POST" name="ballad-form-start">
    <button action="submit">Begin</button>
    </form>
    """
    ).strip()

    async def get(self, request):
        page = Page()
        page = self.compose(request, page)
        return HTMLResponse(page.html)

    @staticmethod
    def render_css_links(request, assets: Grouping[str, list[Loader.Asset]]) -> Generator[str]:
        for asset in assets["text/css"]:
            yield f'<link rel="stylesheet" href="/static/{asset.path.name}" />'

    @staticmethod
    def render_js_links(request, assets: Grouping[str, list[Loader.Asset]]) -> Generator[str]:
        for asset in assets["application/javascript"]:
            yield f'<script src="/static/{asset.path.name}"></script>'

    def compose(self, request, page: Page, story: StoryBuilder=None, turn: StoryBuilder.Turn=None) -> Page:
        assets = getattr(self, "assets", Grouping())

        page.paste(page.zone.meta, self.meta)
        page.paste(page.zone.css, *[line for line in self.render_css_links(request, assets)])
        page.paste(page.zone.body, self.body)
        return page


class Start(HTTPEndpoint):
    async def post(self, request):
        state = request.app.state
        story = state.builder(state.config, assets=getattr(self, "assets", []))
        state.sessions[story.uid] = story
        return RedirectResponse(
            url=request.url_for("session", session_id=story.uid), status_code=303
        )


class Session(HTTPEndpoint):

    async def get(self, request):
        session_id = request.path_params["session_id"]
        state = request.app.state
        story = state.sessions[session_id]

        page = Page()
        with story.turn() as turn:
            if not turn.blocks:
                warnings.warn(f"Unable to cast {story.context.ensemble}")
                return RedirectResponse(url=request.url_for("home"), status_code=300)

            page = self.compose(request, page, story, turn)

        return HTMLResponse(page.html)

    def render_refresh(self, url, notes: dict = {}) -> str:
        try:
            wait = notes.get("delay", 0) + notes.get("offer", 0)
            return f'<meta http-equiv="refresh" content="{wait:.2f};{url}">'
        except TypeError:
            return ""

    def render_inputs_to_command(self, request, story):
        options = story.context.options(story.context.ensemble)
        url = request.url_for("command", session_id=story.uid)
        return textwrap.dedent(
            f"""
            <form role="form" action="{url}" method="post" name="ballad-control-text">
            <fieldset>
            <label for="input-command-text" id="input-command-text-tip">&gt;</label>

            <input
            name="text"
            placeholder="{story.context.prompt}"
            pattern="[\w ]+"
            autofocus="autofocus"
            type="text"
            title="&gt;"
            />
            <button type="submit">Enter</button>
            </fieldset>
            </form>
        """
        )

    def compose(self, request, page: Page, story: StoryBuilder=None, turn: StoryBuilder.Turn=None) -> Page:
        assets = getattr(self, "assets", Grouping())

        try:
            page.paste(page.zone.title, f"<title>{story.title}</title>")
        except AttributeError:
            page.paste(page.zone.title, "<title>Story</title>")

        page.paste(page.zone.meta, Home.meta)
        page.paste(page.zone.css, *sorted(line for line in Home.render_css_links(request, assets)))
        page.paste(page.zone.script, *sorted(line for line in Home.render_js_links(request, assets)))

        html5 = "\n".join(turn.blocks)
        page.paste(page.zone.body, html5)

        offer = story.notes and story.notes[-1]["offer"]
        if offer:
            page.paste(page.zone.meta, self.render_refresh(request.url, story.notes[-1]))
        else:
            page.paste(page.zone.inputs, self.render_inputs_to_command(request, story))

        return page


class Command(HTTPEndpoint):
    async def post(self, request):
        session_id = request.path_params["session_id"]
        state = request.app.state
        story = state.sessions[session_id]

        async with request.form() as form:
            command = form["text"]

        story.action(command)

        return RedirectResponse(
            url=request.url_for("session", session_id=story.uid), status_code=303
        )


class Assembly(HTTPEndpoint):

    class EntitySerializer(JSONResponse):
        encoder = Entity.Encoder()

        def render(self, content) -> bytes:
            return self.encoder.encode(content).encode("utf-8")


    async def get(self, request):
        session_id = request.path_params["session_id"]
        state = request.app.state
        story = state.sessions[session_id]
        assembly = dict(ensemble=story.context.ensemble)
        return self.EntitySerializer(assembly)


async def app_factory(
    assets: list = [],
    builder: StoryBuilder = None,
    config: dict = None,
    routes: list = None,
    static: pathlib.Path = None,
    loop=None,
    **kwargs,
):
    endpoints = dict(
        home=next(reversed(Home.__subclasses__()), Home),
        about=next(reversed(About.__subclasses__()), About),
        start=next(reversed(Start.__subclasses__()), Start),
        session=next(reversed(Session.__subclasses__()), Session),
        assembly=next(reversed(Assembly.__subclasses__()), Assembly),
        command=next(reversed(Command.__subclasses__()), Command),
        **kwargs
    )
    for endpt in endpoints.values():
        endpt.assets = assets.copy()

    routes = routes or [
        Route("/", endpoints["home"], name="home"),
        Route("/about", endpoints["about"]),
        Route("/sessions", endpoints["start"], methods=["POST"], name="start"),
        Route("/session/{session_id:uuid}", endpoints["session"], name="session"),
        Route("/session/{session_id:uuid}/assembly", endpoints["assembly"], name="assembly"),
        Route("/session/{session_id:uuid}/command", endpoints["command"], methods=["POST"], name="command"),
    ]
    if static:
        routes.append(Mount("/static", app=StaticFiles(directory=static), name="static"))

    app = Starlette(routes=routes)
    app.state.builder = builder or next(reversed(StoryBuilder.__subclasses__()), StoryBuilder)
    app.state.config = config
    app.state.sessions = {}

    return app


def quick_start(module, resource=""):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    assets = Grouping.typewise(Loader.discover(module, resource))
    app = loop.run_until_complete(
        app_factory(assets=assets, static=assets.all[0].path.parent, loop=loop)
    )
    settings = hypercorn.Config.from_mapping({"bind": "localhost:8080", "errorlog": "-"})

    loop.run_until_complete(serve(app, settings))
